@@ -16,27 +16,21 @@
 
 #define RFC3244_VERSION 0xff80
 
-krb5_error_code
-process_chpw_request(context, server_handle, realm, keytab,
-                     local_faddr, remote_faddr, req, rep)
-    krb5_context context;
-    void *server_handle;
-    char *realm;
-    krb5_keytab keytab;
-    krb5_fulladdr *local_faddr;
-    krb5_fulladdr *remote_faddr;
-    krb5_data *req;
-    krb5_data *rep;
+static krb5_error_code
+process_chpw_request(krb5_context context, void *server_handle, char *realm,
+                     krb5_keytab keytab, const krb5_fulladdr *local_faddr,
+                     const krb5_fulladdr *remote_faddr, krb5_data *req,
+                     krb5_data *rep)
 {
     krb5_error_code ret;
     char *ptr;
-    int plen, vno;
-    krb5_data ap_req, ap_rep;
-    krb5_auth_context auth_context;
-    krb5_principal changepw;
+    unsigned int plen, vno;
+    krb5_data ap_req, ap_rep = empty_data();
+    krb5_data cipher = empty_data(), clear = empty_data();
+    krb5_auth_context auth_context = NULL;
+    krb5_principal changepw = NULL;
     krb5_principal client, target = NULL;
-    krb5_ticket *ticket;
-    krb5_data cipher, clear;
+    krb5_ticket *ticket = NULL;
     krb5_replay_data replay;
     krb5_error krberror;
     int numresult;
@@ -50,16 +44,7 @@ process_chpw_request(context, server_handle, realm, keytab,
     char addrbuf[100];
     krb5_address *addr = remote_faddr->address;
 
-    ret = 0;
-    rep->length = 0;
-    rep->data = NULL;
-
-    auth_context = NULL;
-    changepw = NULL;
-    ap_rep.length = 0;
-    ticket = NULL;
-    clear.length = 0;
-    cipher.length = 0;
+    *rep = empty_data();
 
     if (req->length < 4) {
         /* either this, or the server is printing bad messages,
@@ -152,29 +137,6 @@ process_chpw_request(context, server_handle, realm, keytab,
         goto chpwfail;
     }
 
-    /* mk_priv requires that the local address be set.
-       getsockname is used for this.  rd_priv requires that the
-       remote address be set.  recvfrom is used for this.  If
-       rd_priv is given a local address, and the message has the
-       recipient addr in it, this will be checked.  However, there
-       is simply no way to know ahead of time what address the
-       message will be delivered *to*.  Therefore, it is important
-       that either no recipient address is in the messages when
-       mk_priv is called, or that no local address is passed to
-       rd_priv.  Both is a better idea, and I have done that.  In
-       summary, when mk_priv is called, *only* a local address is
-       specified.  when rd_priv is called, *only* a remote address
-       is specified.  Are we having fun yet?  */
-
-    ret = krb5_auth_con_setaddrs(context, auth_context, NULL,
-                                 remote_faddr->address);
-    if (ret) {
-        numresult = KRB5_KPASSWD_HARDERROR;
-        strlcpy(strresult, "Failed storing client internet address",
-                sizeof(strresult));
-        goto chpwfail;
-    }
-
     /* construct the ap-rep */
 
     ret = krb5_mk_rep(context, auth_context, &ap_rep);
@@ -190,6 +152,14 @@ process_chpw_request(context, server_handle, realm, keytab,
     cipher.length = (req->data + req->length) - ptr;
     cipher.data = ptr;
 
+    /*
+     * Don't set a remote address in auth_context before calling krb5_rd_priv,
+     * so that we can work against clients behind a NAT.  Reflection attacks
+     * aren't a concern since we use sequence numbers and since our requests
+     * don't look anything like our responses.  Also don't set a local address,
+     * since we don't know what interface the request was received on.
+     */
+
     ret = krb5_rd_priv(context, auth_context, &cipher, &clear, &replay);
     if (ret) {
         numresult = KRB5_KPASSWD_HARDERROR;
@@ -211,8 +181,7 @@ process_chpw_request(context, server_handle, realm, keytab,
             goto chpwfail;
         }
 
-        memset(clear.data, 0, clear.length);
-        free(clear.data);
+        zapfree(clear.data, clear.length);
 
         clear = *clear_data;
         free(clear_data);
@@ -258,11 +227,9 @@ process_chpw_request(context, server_handle, realm, keytab,
         errmsg = krb5_get_error_message(context, ret);
 
     /* zap the password */
-    memset(clear.data, 0, clear.length);
-    memset(ptr, 0, clear.length);
-    free(clear.data);
-    free(ptr);
-    clear.length = 0;
+    zapfree(clear.data, clear.length);
+    zapfree(ptr, clear.length);
+    clear = empty_data();
 
     clen = strlen(clientstr);
     trunc_name(&clen, &cdots);
@@ -277,6 +244,7 @@ process_chpw_request(context, server_handle, realm, keytab,
         salen = sizeof(*sin);
         break;
     }
+#ifdef KRB5_USE_INET6
     case ADDRTYPE_INET6: {
         struct sockaddr_in6 *sin6 = ss2sin6(&ss);
 
@@ -286,6 +254,7 @@ process_chpw_request(context, server_handle, realm, keytab,
         salen = sizeof(*sin6);
         break;
     }
+#endif
     default: {
         struct sockaddr *sa = ss2sa(&ss);
 
@@ -315,15 +284,13 @@ process_chpw_request(context, server_handle, realm, keytab,
             targetp = clientstr;
         }
 
-        krb5_klog_syslog(LOG_NOTICE, "setpw request from %s by %.*s%s for %.*s%s: %s",
-                         addrbuf,
-                         (int) clen, clientstr, cdots,
-                         (int) tlen, targetp, tdots,
+        krb5_klog_syslog(LOG_NOTICE, _("setpw request from %s by %.*s%s for "
+                                       "%.*s%s: %s"), addrbuf, (int) clen,
+                         clientstr, cdots, (int) tlen, targetp, tdots,
                          errmsg ? errmsg : "success");
     } else {
-        krb5_klog_syslog(LOG_NOTICE, "chpw request from %s for %.*s%s: %s",
-                         addrbuf,
-                         (int) clen, clientstr, cdots,
+        krb5_klog_syslog(LOG_NOTICE, _("chpw request from %s for %.*s%s: %s"),
+                         addrbuf, (int) clen, clientstr, cdots,
                          errmsg ? errmsg : "success");
     }
     switch (ret) {
@@ -359,7 +326,7 @@ chpwfail:
 
     memcpy(ptr, strresult, strlen(strresult));
 
-    cipher.length = 0;
+    cipher = empty_data();
 
     if (ap_rep.length) {
         ret = krb5_auth_con_setaddrs(context, auth_context,
@@ -389,7 +356,7 @@ chpwfail:
 
         if (ap_rep.length) {
             free(ap_rep.data);
-            ap_rep.length = 0;
+            ap_rep = empty_data();
         }
 
         krberror.ctime = 0;
@@ -426,13 +393,9 @@ chpwfail:
 
     /* construct the reply */
 
-    rep->length = 6 + ap_rep.length + cipher.length;
-    rep->data = (char *) malloc(rep->length);
-    if (rep->data == NULL) {
-        rep->length = 0;        /* checked by caller */
-        ret = ENOMEM;
+    ret = alloc_data(rep, 6 + ap_rep.length + cipher.length);
+    if (ret)
         goto bailout;
-    }
     ptr = rep->data;
 
     /* length */
@@ -462,48 +425,36 @@ chpwfail:
     memcpy(ptr, cipher.data, cipher.length);
 
 bailout:
-    if (auth_context)
-        krb5_auth_con_free(context, auth_context);
-    if (changepw)
-        krb5_free_principal(context, changepw);
-    if (ap_rep.length)
-        free(ap_rep.data);
-    if (ticket)
-        krb5_free_ticket(context, ticket);
-    if (clear.length)
-        free(clear.data);
-    if (cipher.length)
-        free(cipher.data);
-    if (target)
-        krb5_free_principal(context, target);
-    if (targetstr)
-        krb5_free_unparsed_name(context, targetstr);
-    if (clientstr)
-        krb5_free_unparsed_name(context, clientstr);
-    if (errmsg)
-        krb5_free_error_message(context, errmsg);
-
-    return(ret);
+    krb5_auth_con_free(context, auth_context);
+    krb5_free_principal(context, changepw);
+    krb5_free_ticket(context, ticket);
+    free(ap_rep.data);
+    free(clear.data);
+    free(cipher.data);
+    krb5_free_principal(context, target);
+    krb5_free_unparsed_name(context, targetstr);
+    krb5_free_unparsed_name(context, clientstr);
+    krb5_free_error_message(context, errmsg);
+    return ret;
 }
 
 /* Dispatch routine for set/change password */
-krb5_error_code
-dispatch(void *handle,
-         struct sockaddr *local_saddr, const krb5_fulladdr *remote_faddr,
-         krb5_data *request, krb5_data **response, int is_tcp)
+void
+dispatch(void *handle, struct sockaddr *local_saddr,
+         const krb5_fulladdr *remote_faddr, krb5_data *request, int is_tcp,
+         verto_ctx *vctx, loop_respond_fn respond, void *arg)
 {
     krb5_error_code ret;
     krb5_keytab kt = NULL;
     kadm5_server_handle_t server_handle = (kadm5_server_handle_t)handle;
     krb5_fulladdr local_faddr;
     krb5_address **local_kaddrs = NULL, local_kaddr_buf;
-
-    *response = NULL;
+    krb5_data *response = NULL;
 
     if (local_saddr == NULL) {
         ret = krb5_os_localaddr(server_handle->context, &local_kaddrs);
         if (ret != 0)
-            goto cleanup;
+            goto egress;
 
         local_faddr.address = local_kaddrs[0];
         local_faddr.port = 0;
@@ -514,16 +465,14 @@ dispatch(void *handle,
 
     ret = krb5_kt_resolve(server_handle->context, "KDB:", &kt);
     if (ret != 0) {
-        krb5_klog_syslog(LOG_ERR, "chpw: Couldn't open admin keytab %s",
+        krb5_klog_syslog(LOG_ERR, _("chpw: Couldn't open admin keytab %s"),
                          krb5_get_error_message(server_handle->context, ret));
-        goto cleanup;
+        goto egress;
     }
 
-    *response = (krb5_data *)malloc(sizeof(krb5_data));
-    if (*response == NULL) {
-        ret = ENOMEM;
-        goto cleanup;
-    }
+    response = k5alloc(sizeof(krb5_data), &ret);
+    if (response == NULL)
+        goto egress;
 
     ret = process_chpw_request(server_handle->context,
                                handle,
@@ -532,17 +481,11 @@ dispatch(void *handle,
                                &local_faddr,
                                remote_faddr,
                                request,
-                               *response);
-
-cleanup:
-    if (local_kaddrs != NULL)
-        krb5_free_addresses(server_handle->context, local_kaddrs);
-
-    if ((*response)->data == NULL) {
-        free(*response);
-        *response = NULL;
-    }
+                               response);
+egress:
+    if (ret)
+        krb5_free_data(server_handle->context, response);
+    krb5_free_addresses(server_handle->context, local_kaddrs);
     krb5_kt_close(server_handle->context, kt);
-
-    return ret;
+    (*respond)(arg, ret, ret == 0 ? response : NULL);
 }
